"""
Copyright 2021 Objectiv B.V.
"""
from sklearn.linear_model import LogisticRegression as LogisticRegression_sk  # type: ignore
from modelhub.metrics import Metrics
from typing import TYPE_CHECKING
from bach.series import SeriesBoolean

if TYPE_CHECKING:
    from bach.dataframe import DataFrame
    from bach.series import SeriesFloat64


class LogisticRegression:
    """
    A binary class logistic regression model that takes Bach data objects as input data.

    It is based on sklearn's LogisticRegression. All supported methods are run on the database using Bach
    directly, with the exeption of :py:meth:`fit`. For the `fit` method,
    data is extracted from the database before applying sklearn's fit method on the data.

    All parameters to instantiate sklearn's logistic regression are supported.

    For the full documentation, including a description of the parameters, of the Logistic Regression
    model see:
    https://scikit-learn.org/stable/modules/generated/sklearn.linear_model.LogisticRegression.html
    """
    def __init__(self, *args, **kwargs):
        self._model = LogisticRegression_sk(*args, **kwargs)

    def __getattr__(self, item):
        sklearn_parameters = ['C', 'class_weight', 'dual', 'fit_intercept', 'intercept_scaling', 'l1_ratio',
                              'max_iter', 'multi_class', 'n_jobs', 'penalty', 'random_state', 'solver', 'tol',
                              'verbose', 'warm_start']
        sklearn_attributes = ['classes_', 'coef_', 'intercept_', 'n_features_in_', 'feature_names_in_',
                              'n_iter_']
        if item in sklearn_parameters + sklearn_attributes:
            return getattr(self._model, item)
        raise AttributeError(f"no attribute '{item}'")

    def __repr__(self):
        return self._model.__repr__()

    def _decision_function(self, X: 'DataFrame'):
        if len(X.data_columns) != len(self._model.coef_[0]):
            raise ValueError("incorrect number of columns in X")
        X_copy = X.copy()
        X_copy['confidence_score'] = self._model.intercept_[0]
        for column, coef in zip(X.data_columns, self._model.coef_[0]):
            X_copy['confidence_score'] = X_copy['confidence_score'] + X_copy[column] * coef

        return X_copy['confidence_score']

    def fit(self, X: 'DataFrame', y: 'SeriesBoolean'):
        """
        Fits a binary class logistic regression model.

        .. important::
            This method uses sklearns LogisticRegression.fit, meaning that the data in the database gets
            exported first before fitting the data.

        :param X: DataFrame with features.
        :param y: Series with the target variable.
        """

        if not isinstance(y, SeriesBoolean):
            raise TypeError(f"y is of type {type(y)}, should be SeriesBoolean")

        data = X.copy()
        data[y.name] = y
        pdf = data.to_pandas()
        X_p = pdf[X.data_columns]
        y_p = pdf[y.name]

        return self._model.fit(X_p, y_p)

    def predict(self, X: 'DataFrame') -> SeriesBoolean:
        """
        Predict the labels based on the fitted estimator.

        :param X: DataFrame with the same features as the training data set.
        """
        series = self.predict_proba(X) > .5
        return series.copy_override(name='labels')

    def predict_proba(self, X: 'DataFrame') -> 'SeriesFloat64':
        """
        Predict the probability of the 'True' label based on the fitted estimator.

        :param X: DataFrame with the same features as the training data set.
        """
        confidence_score = self._decision_function(X)
        probability = confidence_score.exp() / (confidence_score.exp() + 1.)
        return probability.copy_override(name='probability')

    def score(self, X: 'DataFrame', y: 'SeriesBoolean') -> float:
        """
        Returns the accuracy score from a series of true values compared to predicted values.

        :param X: DataFrame with the same features as the training data set.
        :param y: Series with the true target variable.
        :returns: a single value with the proportion of correct predicted labels.

        .. note::
            This function queries the database.
        """
        y_pred = self.predict(X)
        return Metrics.accuracy_score(y, y_pred)

    def set_params(self, **params):
        """
        Set parameters for the model. Use :py:meth:`get_params` to see the available parameters that can be
        set.
        """
        self._model.set_params(**params)

    def get_params(self) -> dict:
        """
        Return the parameters as a dictionary.
        """
        return self._model.get_params()



"""
Copyright 2021 Objectiv B.V.
"""

# Any import from modelhub initializes all the types, do not remove
from modelhub import __version__
from tests_modelhub.data_and_utils.utils import get_objectiv_dataframe_test
from tests.functional.bach.test_data_and_utils import assert_equals_data


def test_frequency(db_params):
    df, modelhub = get_objectiv_dataframe_test(db_params)
    s = modelhub.aggregate.frequency(df)

    assert_equals_data(
        s,
        expected_columns=['session_id_nunique', 'user_id_nunique'],
        expected_data=[
            [1, 1],
            [2, 3]
        ]
    )



"""
Copyright 2021 Objectiv B.V.
"""

import pytest
# Any import from modelhub initializes all the types, do not remove
from modelhub import __version__
from tests_modelhub.data_and_utils.utils import get_objectiv_dataframe_test
from tests.functional.bach.test_data_and_utils import assert_equals_data
from uuid import UUID


def test_defaults(db_params):
    # setting nothing, thus using all defaults (which is just moment without formatting)
    df, modelhub = get_objectiv_dataframe_test(db_params)
    s = modelhub.aggregate.unique_sessions(df)

    assert_equals_data(
        s,
        expected_columns=['time_aggregation', 'unique_sessions'],
        expected_data=[
            ['2021-11-29 10:23:36.286000', 1],
            ['2021-11-29 10:23:36.287000', 1],
            ['2021-11-30 10:23:36.267000', 1],
            ['2021-11-30 10:23:36.287000', 1],
            ['2021-11-30 10:23:36.290000', 1],
            ['2021-11-30 10:23:36.291000', 1],
            ['2021-12-01 10:23:36.276000', 1],
            ['2021-12-01 10:23:36.279000', 1],
            ['2021-12-02 10:23:36.281000', 1],
            ['2021-12-02 14:23:36.282000', 1],
            ['2021-12-03 10:23:36.283000', 1]]
    )


def test_no_grouping(db_params):
    # not grouping to anything
    df, modelhub = get_objectiv_dataframe_test(db_params)
    s = modelhub.aggregate.unique_sessions(df, groupby=None)

    assert_equals_data(
        s,
        expected_columns=['unique_sessions'],
        expected_data=[
            [7]
        ]
    )


def test_time_aggregation_in_df(db_params):
    # using time_aggregation (and default groupby: mh.time_agg(df, ))
    df, modelhub = get_objectiv_dataframe_test(db_params, time_aggregation='%Y-%m-%d')
    s = modelhub.aggregate.unique_sessions(df)

    assert_equals_data(
        s,
        expected_columns=['time_aggregation', 'unique_sessions'],
        expected_data=[['2021-11-29', 1],
            ['2021-11-30', 2],
            ['2021-12-01', 1],
            ['2021-12-02', 2],
            ['2021-12-03', 1]]
    )


def test_overriding_time_aggregation_in(db_params):
    # overriding time_aggregation
    df, modelhub = get_objectiv_dataframe_test(db_params, time_aggregation='%Y-%m-%d')
    s = modelhub.aggregate.unique_sessions(df, groupby=modelhub.time_agg(df, '%Y-%m'))

    assert_equals_data(
        s,
        expected_columns=['time_aggregation', 'unique_sessions'],
        expected_data=[['2021-11', 3],
            ['2021-12', 4]]
    )


def test_groupby(db_params):
    # group by other columns
    df, modelhub = get_objectiv_dataframe_test(db_params, time_aggregation='%Y-%m-%d')
    s = modelhub.aggregate.unique_sessions(df, groupby='event_type')

    assert_equals_data(
        s,
        expected_columns=['event_type', 'unique_sessions'],
        expected_data=[['ClickEvent', 7]]
    )


def test_groupby_incl_time_agg(db_params):
    # group by other columns (as series), including time_agg
    df, modelhub = get_objectiv_dataframe_test(db_params, time_aggregation='%Y-%m-%d')
    s = modelhub.aggregate.unique_sessions(df, groupby=[modelhub.time_agg(df, '%Y-%m'), df.user_id])

    assert_equals_data(
        s,
        expected_columns=['time_aggregation', 'user_id', 'unique_sessions'],
        expected_data=[
            ['2021-11', UUID('b2df75d2-d7ca-48ac-9747-af47d7a4a2b1'), 1],
            ['2021-11', UUID('b2df75d2-d7ca-48ac-9747-af47d7a4a2b2'), 2],
            ['2021-12', UUID('b2df75d2-d7ca-48ac-9747-af47d7a4a2b1'), 1],
            ['2021-12', UUID('b2df75d2-d7ca-48ac-9747-af47d7a4a2b3'), 2],
            ['2021-12', UUID('b2df75d2-d7ca-48ac-9747-af47d7a4a2b4'), 1]
        ],
        convert_uuid=True,
    )


def test_groupby_illegal_column(db_params):
    # include column that is used for grouping in groupby
    df, modelhub = get_objectiv_dataframe_test(db_params, time_aggregation='%Y-%m-%d')
    with pytest.raises(KeyError, match='is in groupby but is needed for aggregation: not allowed to '
                                         'group on that'):
        modelhub.aggregate.unique_sessions(df, groupby=[modelhub.time_agg(df, '%Y-%m'), df.session_id])



"""
Copyright 2021 Objectiv B.V.
"""
import pytest
from modelhub import __version__  # Any import from modelhub initializes all the types, do not remove
from tests_modelhub.data_and_utils.utils import get_df_with_json_data_real, DBParams
from tests.functional.bach.test_data_and_utils import assert_equals_data


def test_get_real_data(db_params: DBParams):
    bt = get_df_with_json_data_real(db_params)
    assert_equals_data(
        bt,
        use_to_pandas=True,
        expected_columns=['_index_event_id', 'event_id', 'global_contexts', 'location_stack'],
        expected_data=[
            [1, 1, [{'id': 'rod-web-demo', '_type': 'ApplicationContext'}, {'id': 'http_context', '_type': 'HttpContext', 'referrer': 'https://rick.objectiv.io/', 'user_agent': 'Mozilla/5.0 (X11; Ubuntu; Linux x86_64; rv:94.0) Gecko/20100101 Firefox/94.0', 'remote_address': '144.144.144.144'}, {'id': 'f84446c6-eb76-4458-8ef4-93ade596fd5b', '_type': 'CookieIdContext', 'cookie_id': 'f84446c6-eb76-4458-8ef4-93ade596fd5b'}], [{'id': '#document', 'url': 'https://rick.objectiv.io/', '_type': 'WebDocumentContext'}, {'id': 'home', '_type': 'SectionContext'}, {'id': 'yep', '_type': 'SectionContext'}, {'id': 'cc91EfoBh8A', '_type': 'SectionContext'}]],
            [2, 2, [{'id': 'rod-web-demo', '_type': 'ApplicationContext'}, {'id': 'http_context', '_type': 'HttpContext', 'referrer': 'https://rick.objectiv.io/', 'user_agent': 'Mozilla/5.0 (X11; Ubuntu; Linux x86_64; rv:94.0) Gecko/20100101 Firefox/94.0', 'remote_address': '144.144.144.144'}, {'id': 'f84446c6-eb76-4458-8ef4-93ade596fd5b', '_type': 'CookieIdContext', 'cookie_id': 'f84446c6-eb76-4458-8ef4-93ade596fd5b'}], [{'id': '#document', 'url': 'https://rick.objectiv.io/', '_type': 'WebDocumentContext', '_types': ['AbstractContext', 'AbstractLocationContext', 'SectionContext', 'WebDocumentContext']}, {'id': 'navigation', '_type': 'NavigationContext', '_types': ['AbstractContext', 'AbstractLocationContext', 'NavigationContext', 'SectionContext']}]],
            [3, 3, [{'id': 'rod-web-demo', '_type': 'ApplicationContext'}, {'id': 'http_context', '_type': 'HttpContext', 'referrer': 'https://rick.objectiv.io/', 'user_agent': 'Mozilla/5.0 (X11; Ubuntu; Linux x86_64; rv:94.0) Gecko/20100101 Firefox/94.0', 'remote_address': '144.144.144.144'}, {'id': 'f84446c6-eb76-4458-8ef4-93ade596fd5b', '_type': 'CookieIdContext', 'cookie_id': 'f84446c6-eb76-4458-8ef4-93ade596fd5b'}], [{'id': '#document', 'url': 'https://rick.objectiv.io/', '_type': 'WebDocumentContext'}, {'id': 'home', '_type': 'SectionContext'}, {'id': 'new', '_type': 'SectionContext'}, {'id': 'BeyEGebJ1l4', '_type': 'SectionContext'}]],
            [4, 4, [{'id': 'rod-web-demo', '_type': 'ApplicationContext'}, {'id': 'http_context', '_type': 'HttpContext', 'referrer': 'https://rick.objectiv.io/', 'user_agent': 'Mozilla/5.0 (X11; Ubuntu; Linux x86_64; rv:94.0) Gecko/20100101 Firefox/94.0', 'remote_address': '144.144.144.144'}, {'id': 'f84446c6-eb76-4458-8ef4-93ade596fd5b', '_type': 'CookieIdContext', 'cookie_id': 'f84446c6-eb76-4458-8ef4-93ade596fd5b'}], [{'id': '#document', 'url': 'https://rick.objectiv.io/', '_type': 'WebDocumentContext'}, {'id': 'home', '_type': 'SectionContext'}, {'id': 'new', '_type': 'SectionContext'}, {'id': 'yBwD4iYcWC4', '_type': 'SectionContext'}]],
            [5, 5, [{'id': 'rod-web-demo', '_type': 'ApplicationContext'}, {'id': 'http_context', '_type': 'HttpContext', 'referrer': 'https://rick.objectiv.io/', 'user_agent': 'Mozilla/5.0 (X11; Ubuntu; Linux x86_64; rv:94.0) Gecko/20100101 Firefox/94.0', 'remote_address': '144.144.144.144'}, {'id': 'f84446c6-eb76-4458-8ef4-93ade596fd5b', '_type': 'CookieIdContext', 'cookie_id': 'f84446c6-eb76-4458-8ef4-93ade596fd5b'}], [{'id': '#document', 'url': 'https://rick.objectiv.io/', '_type': 'WebDocumentContext'}, {'id': 'home', '_type': 'SectionContext'}, {'id': 'new', '_type': 'SectionContext'}, {'id': 'eYuUAGXN0KM', '_type': 'MediaPlayerContext'}]],
            [6, 6, [{'id': 'rod-web-demo', '_type': 'ApplicationContext'}, {'id': 'http_context', '_type': 'HttpContext', 'referrer': 'https://rick.objectiv.io/', 'user_agent': 'Mozilla/5.0 (X11; Ubuntu; Linux x86_64; rv:94.0) Gecko/20100101 Firefox/94.0', 'remote_address': '144.144.144.144'}, {'id': 'f84446c6-eb76-4458-8ef4-93ade596fd5b', '_type': 'CookieIdContext', 'cookie_id': 'f84446c6-eb76-4458-8ef4-93ade596fd5b'}], [{'id': '#document', 'url': 'https://rick.objectiv.io/', '_type': 'WebDocumentContext'}]]
        ]
    )


def test_objectiv_stack_type(db_params):
    bt = get_df_with_json_data_real(db_params)

    bt['a'] = bt.global_contexts.astype('objectiv_global_context')
    bts = bt.a.objectiv.get_from_context_with_type_series("CookieIdContext", "cookie_id")
    assert_equals_data(
        bts,
        expected_columns=['_index_event_id', 'a'],
        expected_data=[
            [1, 'f84446c6-eb76-4458-8ef4-93ade596fd5b'],
            [2, 'f84446c6-eb76-4458-8ef4-93ade596fd5b'],
            [3, 'f84446c6-eb76-4458-8ef4-93ade596fd5b'],
            [4, 'f84446c6-eb76-4458-8ef4-93ade596fd5b'],
            [5, 'f84446c6-eb76-4458-8ef4-93ade596fd5b'],
            [6, 'f84446c6-eb76-4458-8ef4-93ade596fd5b']
        ]
    )


def test_objectiv_stack_type2(db_params):
    bt = get_df_with_json_data_real(db_params)

    bt['a'] = bt.global_contexts.astype('objectiv_global_context')
    bts = bt.a.global_contexts.user_agent
    assert_equals_data(
        bts,
        expected_columns=['_index_event_id', 'a'],
        expected_data=[
            [1, 'Mozilla/5.0 (X11; Ubuntu; Linux x86_64; rv:94.0) Gecko/20100101 Firefox/94.0'],
            [2, 'Mozilla/5.0 (X11; Ubuntu; Linux x86_64; rv:94.0) Gecko/20100101 Firefox/94.0'],
            [3, 'Mozilla/5.0 (X11; Ubuntu; Linux x86_64; rv:94.0) Gecko/20100101 Firefox/94.0'],
            [4, 'Mozilla/5.0 (X11; Ubuntu; Linux x86_64; rv:94.0) Gecko/20100101 Firefox/94.0'],
            [5, 'Mozilla/5.0 (X11; Ubuntu; Linux x86_64; rv:94.0) Gecko/20100101 Firefox/94.0'],
            [6, 'Mozilla/5.0 (X11; Ubuntu; Linux x86_64; rv:94.0) Gecko/20100101 Firefox/94.0']
        ]
    )


def test_objectiv_stack_type3(db_params):
    bt = get_df_with_json_data_real(db_params)

    bt['b'] = bt.location_stack.astype('objectiv_location_stack')
    bts = bt.b.location_stack.navigation_features
    assert_equals_data(
        bts,
        expected_columns=['_index_event_id', 'b'],
        expected_data=[
            [1, []],
            [2, [{'id': 'navigation', '_type': 'NavigationContext', '_types': ['AbstractContext', 'AbstractLocationContext', 'NavigationContext', 'SectionContext']}]],
            [3, []],
            [4, []],
            [5, []],
            [6, []]
        ],
        use_to_pandas=True,
    )


def test_objectiv_stack_type4(db_params):
    bt = get_df_with_json_data_real(db_params)

    bt['b'] = bt.location_stack.astype('objectiv_location_stack')
    bts = bt.b.location_stack.feature_stack
    assert_equals_data(
        bts,
        use_to_pandas=True,
        expected_columns=['_index_event_id', 'b'],
        expected_data=[
            [1, [{'id': '#document', '_type': 'WebDocumentContext'}, {'id': 'home', '_type': 'SectionContext'}, {'id': 'yep', '_type': 'SectionContext'}, {'id': 'cc91EfoBh8A', '_type': 'SectionContext'}]],
            [2, [{'id': '#document', '_type': 'WebDocumentContext'}, {'id': 'navigation', '_type': 'NavigationContext'}]],
            [3, [{'id': '#document', '_type': 'WebDocumentContext'}, {'id': 'home', '_type': 'SectionContext'}, {'id': 'new', '_type': 'SectionContext'}, {'id': 'BeyEGebJ1l4', '_type': 'SectionContext'}]],
            [4, [{'id': '#document', '_type': 'WebDocumentContext'}, {'id': 'home', '_type': 'SectionContext'}, {'id': 'new', '_type': 'SectionContext'}, {'id': 'yBwD4iYcWC4', '_type': 'SectionContext'}]],
            [5, [{'id': '#document', '_type': 'WebDocumentContext'}, {'id': 'home', '_type': 'SectionContext'}, {'id': 'new', '_type': 'SectionContext'}, {'id': 'eYuUAGXN0KM', '_type': 'MediaPlayerContext'}]],
            [6, [{'id': '#document', '_type': 'WebDocumentContext'}]]
        ]

    )


def test_objectiv_stack_type5(db_params):
    bt = get_df_with_json_data_real(db_params)

    bt['b'] = bt.location_stack.astype('objectiv_location_stack')
    bts = bt.b.location_stack.nice_name
    assert_equals_data(
        bts,
        expected_columns=['_index_event_id', 'b'],
        expected_data=[
            [1, 'Section: cc91EfoBh8A located at Web Document: #document => Section: home => Section: yep'],
            [2, 'Navigation: navigation located at Web Document: #document'],
            [3, 'Section: BeyEGebJ1l4 located at Web Document: #document => Section: home => Section: new'],
            [4, 'Section: yBwD4iYcWC4 located at Web Document: #document => Section: home => Section: new'],
            [5, 'Media Player: eYuUAGXN0KM located at Web Document: #document => Section: home => Section: new'],
            [6, 'Web Document: #document']
        ]
    )



"""
Copyright 2021 Objectiv B.V.
"""
import bach
from bach.series import Series
from sql_models.constants import NotSet, not_set
from typing import cast, List, Union, TYPE_CHECKING

from modelhub.decorators import use_only_required_objectiv_series

if TYPE_CHECKING:
    from modelhub import ModelHub
    from modelhub.series import SeriesLocationStack

GroupByType = Union[List[Union[str, Series]], str, Series, NotSet]


class Aggregate:
    """
    Models that return aggregated data in some form from the original DataFrame with Objectiv data.
    """

    def __init__(self, mh: 'ModelHub'):
        self._mh = mh

    def _check_groupby(self,
                       data: bach.DataFrame,
                       groupby: Union[List[Union[str, Series]], str, Series],
                       not_allowed_in_groupby: str = None
                       ):

        if data.group_by:
            raise ValueError("can't run model hub models on a grouped DataFrame, please use parameters "
                             "(ie groupby of the model")

        groupby_list = groupby if isinstance(groupby, list) else [groupby]
        groupby_list = [] if groupby is None else groupby_list

        if not_allowed_in_groupby is not None and not_allowed_in_groupby not in data.data_columns:
            raise ValueError(f'{not_allowed_in_groupby} column is required for this model but it is not in '
                             f'the DataFrame')

        if not_allowed_in_groupby:
            for key in groupby_list:
                new_key = data[key] if isinstance(key, str) else key
                if new_key.equals(data[not_allowed_in_groupby]):
                    raise KeyError(f'"{not_allowed_in_groupby}" is in groupby but is needed for aggregation: '
                                   f'not allowed to group on that')

        grouped_data = data.groupby(groupby_list)
        return grouped_data

    def _generic_aggregation(self,
                             data: bach.DataFrame,
                             groupby: Union[List[Union[str, Series]], str, Series],
                             column: str,
                             name: str):

        data = self._check_groupby(data=data,
                                   groupby=groupby,
                                   not_allowed_in_groupby=column)

        series = data[column].nunique()
        return series.copy_override(name=name)

    @use_only_required_objectiv_series(
        required_series=['user_id', 'moment'], include_series_from_params=['groupby'],
    )
    def unique_users(self,
                     data: bach.DataFrame,
                     groupby: GroupByType = not_set) -> bach.SeriesInt64:
        """
        Calculate the unique users in the Objectiv ``data``.

        :param data: :py:class:`bach.DataFrame` to apply the method on.
        :param groupby: sets the column(s) to group by.

            - if not_set it defaults to using :py:attr:`ModelHub.time_agg`.
            - if None it aggregates over all data.
        :returns: series with results.
        """

        groupby = [self._mh.time_agg(data)] if groupby is not_set else groupby

        return self._generic_aggregation(data=data,
                                         groupby=groupby,
                                         column='user_id',
                                         name='unique_users')

    @use_only_required_objectiv_series(
        required_series=['session_id', 'moment'], include_series_from_params=['groupby'],
    )
    def unique_sessions(self,
                        data: bach.DataFrame,
                        groupby: GroupByType = not_set) -> bach.SeriesInt64:
        """
        Calculate the unique sessions in the Objectiv ``data``.

        :param data: :py:class:`bach.DataFrame` to apply the method on.
        :param groupby: sets the column(s) to group by.

            - if not_set it defaults to using :py:attr:`ModelHub.time_agg`.
            - if None it aggregates over all data.
        :returns: series with results.
        """

        groupby = [self._mh.time_agg(data)] if groupby is not_set else groupby

        return self._generic_aggregation(data=data,
                                         groupby=groupby,
                                         column='session_id',
                                         name='unique_sessions')

    @use_only_required_objectiv_series(
        required_series=['session_id', 'moment'], include_series_from_params=['groupby']
    )
    def session_duration(self,
                         data: bach.DataFrame,
                         groupby: GroupByType = not_set,
                         exclude_bounces: bool = True,
                         method: 'str' = 'mean') -> bach.SeriesInt64:
        """
        Calculate the duration of sessions.

        With default `method`, it calculates the mean of the session duration over the `groupby`.

        :param data: :py:class:`bach.DataFrame` to apply the method on.
        :param groupby: sets the column(s) to group by.

            - if not_set it defaults to using :py:attr:`ModelHub.time_agg`.
            - if None it aggregates over all data.
        :param exclude_bounces: if True only session durations greater than 0 will be considered
        :param method: 'mean' or 'sum'
        :returns: series with results.
        """

        if groupby is not_set:
            groupby = self._mh.time_agg(data)

        if groupby is None:
            new_groupby = []
        elif not isinstance(groupby, list):
            new_groupby = [groupby]
        else:
            new_groupby = groupby
        new_groupby.append(data.session_id.copy_override(name='__session_id'))

        gdata = self._check_groupby(data=data, groupby=new_groupby)
        session_duration = gdata.aggregate({'moment': ['min', 'max']})
        session_duration['session_duration'] = session_duration['moment_max'] - session_duration['moment_min']

        if exclude_bounces:
            session_duration = session_duration[(session_duration['session_duration'].dt.total_seconds > 0)]

        if method not in ['sum', 'mean']:
            raise ValueError("only 'sum and 'mean' are supported for `method`")

        grouped_data = session_duration.groupby(session_duration.index_columns[:-1]).session_duration
        if method == 'sum':
            return grouped_data.sum()
        return grouped_data.mean()

    @use_only_required_objectiv_series(required_series=['user_id', 'session_id'])
    def frequency(self, data: bach.DataFrame) -> bach.SeriesInt64:
        """
        Calculate a frequency table for the number of users by number of sessions.

        :param data: :py:class:`bach.DataFrame` to apply the method on.
        :returns: series with results.
        """
        total_sessions_user = data.groupby(['user_id']).aggregate({'session_id': 'nunique'}).reset_index()
        frequency = total_sessions_user.groupby(['session_id_nunique']).aggregate({'user_id': 'nunique'})

        return frequency.user_id_nunique

    @use_only_required_objectiv_series(
        required_series=['global_contexts', 'location_stack', 'user_id', 'stack_event_types', 'event_type'],
    )
    def top_product_features(self,
                             data: bach.DataFrame,
                             location_stack: 'SeriesLocationStack' = None,
                             event_type: str = 'InteractiveEvent') -> bach.DataFrame:
        """
        Calculate the top used features in the product.

        :param data: :py:class:`bach.DataFrame` to apply the method on.
        :param location_stack: the location stack

            - can be any slice of a :py:class:`modelhub.SeriesLocationStack` type column
            - if None - the whole location stack is taken.
        :param event_type: event type. Must be a valid event_type (either parent or child).
        :returns: bach DataFrame with results.
        """

        data = data.copy()

        # the following columns have to be in the data
        data['__application'] = data.global_contexts.gc.application

        if location_stack is not None:
            data['__feature_nice_name'] = location_stack.ls.nice_name
        else:
            data['__feature_nice_name'] = data.location_stack.ls.nice_name

        groupby_col = ['__application', '__feature_nice_name', 'event_type']

        # selects specific event types, so stack_event_types must be a superset of [event_types]
        interactive_events = data[data.stack_event_types.json.array_contains(event_type)]

        # users by feature
        users_feature = interactive_events.groupby(groupby_col).agg({'user_id': 'nunique'})

        # remove double underscores from the columns name
        columns = {
            col: col[2:] if col.startswith('__') else col
            for col in groupby_col
        }
        _index = list(columns.values())
        users_feature = users_feature.reset_index().rename(columns=columns)
        users_feature = users_feature.dropna().set_index(_index)

        return users_feature.sort_values('user_id_nunique', ascending=False)

    @use_only_required_objectiv_series(
        required_series=[
            'global_contexts', 'location_stack', 'user_id', 'stack_event_types',
            # required by Map.conversions_in_time and Map.conversions_counter
            'session_id', 'moment', 'event_type',
        ],
    )
    def top_product_features_before_conversion(self,
                                               data: bach.DataFrame,
                                               name: str,
                                               location_stack: 'SeriesLocationStack' = None,
                                               event_type: str = 'InteractiveEvent') -> bach.DataFrame:
        """
        Calculates what users did before converting by
        combining several models from the model hub.

        :param data: :py:class:`bach.DataFrame` to apply the method on.
        :param name: label of the conversion event.
        :param location_stack: the location stack

            - can be any slice of a :py:class:`modelhub.SeriesLocationStack` type column
            - if None - the whole location stack is taken.
        :param event_type: event type. Must be a valid event_type
            (either parent or child).
        :returns: bach DataFrame with results.
        """

        data = data.copy()

        if not name:
            raise ValueError('Conversion event label is not provided.')

        # temporary workaround, we should not call private constants and methods from Map
        # replace this after having a better solution

        from modelhub.map import _CalculatedConversionSeries
        # get conversion information
        # conversions_counter, is_conversion_event, conversions_in_time
        conversions_df = self._mh.map._get_calculated_conversion_df(
            series_to_calculate=_CalculatedConversionSeries.CONVERSIONS_COUNTER,
            data=data,
            name=name,
            partition='session_id',
        )

        # label sessions with a conversion
        converted_users = conversions_df['__converted'] >= 1

        # label hits where at that point in time, there are 0 conversions in the session
        zero_conversions_at_moment = conversions_df['__conversions_in_time'] == 0
        conversions_df = conversions_df[converted_users & zero_conversions_at_moment]

        # select only user interactions
        data = data[data.stack_event_types.json.array_contains(event_type)]

        # merge with filtered conversion interactive events
        converted_users_filtered = data.merge(conversions_df, on='event_id')
        converted_users_filtered['__application'] = converted_users_filtered.global_contexts.gc.application

        if location_stack is not None:
            converted_users_filtered['__feature_nice_name'] = location_stack.ls.nice_name
        else:
            converted_users_filtered['__feature_nice_name'] = (
                converted_users_filtered.location_stack.ls.nice_name
            )

        converted_users_filtered.materialize(
            node_name='extract_application_and_feature_nice_name', inplace=True,
        )

        groupby_col = ['__application', '__feature_nice_name', 'event_type']
        converted_users_features = self._mh.agg.unique_users(converted_users_filtered,
                                                             groupby=groupby_col)

        # remove double underscores from the columns name
        columns = {
            col: col[2:] if col.startswith('__') else col
            for col in groupby_col
        }
        _index = list(columns.values())
        converted_users_features = converted_users_features.reset_index().rename(columns=columns)
        converted_users_features = converted_users_features.dropna().set_index(_index)

        return converted_users_features.sort_values('unique_users', ascending=False)

    @use_only_required_objectiv_series(required_series=['user_id', 'moment', 'event_type'])
    def retention_matrix(self,
                         data: bach.DataFrame,
                         time_period: str = 'monthly',
                         event_type: str = None,
                         start_date: str = None,
                         end_date: str = None,
                         percentage=False,
                         display=True) -> bach.DataFrame:

        """
        Finds the number of users in a given cohort who are active at a given time
        period, where time is computed with respect to the beginning of each cohort.
        The "active user" is the user who made an action that we are interested in
        that time period.
        Users are divided into mutually exclusive cohorts, which are then
        tracked over time. In our case users are assigned to a cohort based on
        when they made their first action that we are interested in.

        Returns the retention matrix dataframe, it represents users retained across cohorts:

        - index value represents the cohort
        - columns represent the number of given date period since the current cohort
        - values represent number (or percentage) of unique active users of a given cohort

        One can calculate the retention matrix for a given time range, for that
        one can specify start_date a/o end_date.
        N.B. the users' activity starts to be traced from the first date the user is seen in the `data`.


        :param data: :py:class:`bach.DataFrame` to apply the method on.
        :param time_period: can be 'daily', 'weekly', 'monthly' or 'yearly'.
        :param event_type: the event/action that we are interested in.
            Must be a valid event_type (either parent or child).
            if None we take all the events generated by the user.
        :param start_date: start date of the retention matrix, e.g. '2022-04-01'
            if None take all the data.
        :param end_date: end date of the retention matrix, e.g. '2022-05-01'
            if None take all the data.
        :param percentage: if True calculate percentage with respect to the number of a users
            in the cohort, otherwise it leaves the absolute values.
        :param display: if display==True visualize the retention matrix as a heat map

        :returns: retention matrix bach DataFrame.

        .. vimeoplayer::
            :videoid: 723381969
            :trackingid: product-demo-retention-matrix
            :paddingbottom: 58.25%
        """

        available_formats = {'daily', 'weekly', 'monthly', 'yearly'}

        if time_period not in available_formats:
            raise ValueError(f'{time_period} time_period is not available.')

        from datetime import datetime

        _start_date = None
        if start_date is not None:
            try:
                _start_date = datetime.strptime(start_date, '%Y-%m-%d')
            except Exception as e:
                print('Please provide correct start_date formatted as "YYYY-MM-DD".')
                raise e

        _end_date = None
        if end_date is not None:
            try:
                _end_date = datetime.strptime(end_date, '%Y-%m-%d')
            except Exception as e:
                print('Please provide correct end_date formatted as "YYYY-MM-DD".')
                raise e

        data = data.copy()

        # filtering data with the event that we are interested in
        if event_type is not None:
            data = data[data['event_type'] == event_type]

        # for retention matrix calculation we need only event date and user_id
        columns = ['user_id', 'moment']
        data = data[columns]

        # get the first cohort
        cohorts = cast(
            bach.DataFrame, data.groupby('user_id')['moment'].min().reset_index()
        )
        cohorts = cohorts.rename(columns={'moment': 'first_cohort_ts'})

        # add first cohort to our data DataFrame
        data = data.merge(cohorts, on='user_id')

        # help mypy
        from bach import SeriesTimestamp, SeriesTimedelta
        moment = cast(SeriesTimestamp, data['moment'])
        first_cohort_ts = cast(SeriesTimestamp, data['first_cohort_ts'])

        # calculate cohort distance
        if time_period == 'yearly':
            data['cohort'] = moment.dt.strftime('%Y').astype(dtype=int)
            data['first_cohort_yearly'] = first_cohort_ts.dt.strftime('%Y').astype(dtype=int)
            data['cohort_distance'] = data['cohort'] - data['first_cohort_yearly']

            data['first_cohort'] = first_cohort_ts.dt.strftime('%Y')
        elif time_period == 'monthly':
            data['cohort_year'] = moment.dt.strftime('%Y').astype(dtype=int)
            data['first_cohort_year'] = first_cohort_ts.dt.strftime('%Y').astype(dtype=int)
            data['cohort_year_diff'] = data['cohort_year'] - data['first_cohort_year']

            data['cohort_month'] = moment.dt.strftime('%m').astype(dtype=int)
            data['first_cohort_month'] = first_cohort_ts.dt.strftime('%m').astype(dtype=int)
            data['cohort_month_diff'] = data['cohort_month'] - data['first_cohort_month']

            n_months = 12
            data['cohort_distance'] = data['cohort_year_diff'] * n_months + data['cohort_month_diff']

            data['first_cohort'] = first_cohort_ts.dt.strftime('%Y-%m')
        elif time_period == 'weekly':
            n_days = 7.0

            data['cohort'] = moment.dt.date_trunc('week').astype('timestamp')
            data['first_cohort_weekly'] = first_cohort_ts.dt.date_trunc('week')
            cohort_distance = cast(SeriesTimedelta, data['cohort'] - data['first_cohort_weekly'])
            data['cohort_distance'] = cohort_distance.dt.days
            data['cohort_distance'] = data['cohort_distance'] / n_days

            pretty_name = cast(SeriesTimestamp, data['first_cohort_weekly']).dt.strftime('%Y-%m-%d')
            data['first_cohort'] = pretty_name
        else:
            # daily
            data['cohort_distance'] = data['moment'] - data['first_cohort_ts']
            data['cohort_distance'] = cast(SeriesTimedelta, data['cohort_distance']).dt.days
            data['first_cohort'] = first_cohort_ts.dt.strftime('%Y-%m-%d')

        # applying start date filter
        if _start_date is not None:
            data = data[data['first_cohort_ts'] >= _start_date]

        # applying end date filter
        if _end_date is not None:
            data = data[data['moment'] < _end_date]

        # in BigQuery columns cannot start with numbers
        data['cohort_distance_prefix'] = '_'
        data['cohort_distance'] = data['cohort_distance'].astype(dtype=str)
        data['cohort_distance'] = data['cohort_distance_prefix'] + data['cohort_distance']

        retention_matrix = data.groupby(['first_cohort',
                                         'cohort_distance']).agg({'user_id': 'nunique'}).unstack(
            level='cohort_distance')

        # renaming columns, removing string attached after unstacking
        column_name_map = {col: col.replace('__user_id_nunique', '').replace('.0', '')
                           for col in retention_matrix.data_columns}
        retention_matrix = retention_matrix.rename(columns=column_name_map)

        # 'sort' with column names (numerical sorting, even though the columns are strings)
        columns = [f'_{j}' for j in sorted([int(i.replace('_', ''))
                                            for i in retention_matrix.data_columns])]
        retention_matrix = retention_matrix[columns]
        # for BigQuery we need sorting
        retention_matrix = retention_matrix.sort_index()

        if percentage and len(columns):
            first_column = retention_matrix[columns[0]]
            for col in columns:
                retention_matrix[col] = (retention_matrix[col] / first_column) * 100

        if display and len(columns):
            import matplotlib.pyplot as plt
            import seaborn as sns
            fig, ax = plt.subplots(figsize=(20, 8))
            sns.heatmap(retention_matrix.to_pandas(), annot=True, square=True, ax=ax,
                        linewidths=.5, cmap=sns.cubehelix_palette(rot=-.4), fmt='.1f')
            if percentage:
                [t.set_text(f'{t.get_text()}%') for t in ax.texts]

            plt.title('Cohort Analysis')

            nice_name = {
                'daily': 'Days',
                'weekly': 'Weeks',
                'monthly': 'Months',
                'yearly': 'Years'
            }

            plt.xlabel(f'{nice_name[time_period]} After First Event')
            plt.ylabel('First Event Cohort')
            plt.show()

        return retention_matrix



"""
Copyright 2022 Objectiv B.V.
"""
from typing import Dict, Optional

import bach
from modelhub.pipelines.base_pipeline import BaseDataPipeline
from modelhub.util import (
    ObjectivSupportedColumns, get_supported_dtypes_per_objectiv_column, check_objectiv_dataframe
)


class IdentityResolutionPipeline(BaseDataPipeline):
    """
    Pipeline in charge of resolving user identities based on `IdentityContext`.
    This pipeline is dependent on the result from ExtractedContextsPipeline, therefore it expects that
    the result from the latter is generated correctly.


    The steps followed in this pipeline are the following:
        1. _validate_extracted_context_df: Validates if provided DataFrame contains
            user_id, global_contexts and moment series.
        2. _extract_identities_from_global_contexts: Creates new identity (user id) id
            and name values from the first IdentityContext value found for the event
            under the global_contexts series json. Returns a DataFrame considering the last
            registered identity for each user_id. If no identity was found, user_id is not considered.
        3. _resolve_original_user_ids: Replaces original user_ids with the ones extracted from previous step,
            only if an identity was found for it.
        4. _convert_dtypes: Will convert all required identity series to their correct dtype

    If user anonymization is required, call classmethod `anonymize_user_ids_without_identity`. The
        provided dataframe MUST have  `identity_user_id` series.

    Final bach DataFrame will be later validated, it must include:
        - 'identity_user_id', 'user_id', 'global_contexts', 'moment' series.
    """

    IDENTITY_FORMAT = "({}) || '|' || ({})"

    def __init__(self, identity_id: Optional[str] = None):
        self._identity_id = identity_id

    def _get_pipeline_result(
        self,
        extracted_contexts_df: Optional[bach.DataFrame] = None,
        **kwargs,
    ) -> bach.DataFrame:
        """
        Contains steps for solving identities for user_ids in provided dataframe.
        :param extracted_contexts_df: bach DataFrame containing `user_id`, `global_contexts`
            and `moment` series.
        :param identity_id: Identity id to be used for filtering IdentityContexts. If no value is provided,
            all IdentityContexts will be considered.

        returns a bach DataFrame
            - user_id original series dtype will be changed to string
            - identity_user_id series and all series from provided extracted_contexts_df.

        .. note::
            If user has no identity, original value will remain. If anonymization is required, please call
            IdentityResolutionPipeline.anonymize_user_ids_without_identity(df).
        """
        if not extracted_contexts_df:
            raise ValueError(f'{self.__class__.__name__} requires dataframe with extracted contexts.')

        context_df = extracted_contexts_df.copy()

        self._validate_extracted_context_df(context_df)

        context_df[ObjectivSupportedColumns.USER_ID.value] = (
            context_df[ObjectivSupportedColumns.USER_ID.value].astype(bach.SeriesString.dtype)
        )
        identity_context_df = self._extract_identities_from_global_contexts(context_df)

        context_df = self._resolve_original_user_ids(context_df, identity_context_df)
        return self._convert_dtypes(df=context_df)

    @classmethod
    def validate_pipeline_result(cls, result: bach.DataFrame) -> None:
        """
        Checks if we are returning required Objectiv series with respective dtype.
        """
        check_objectiv_dataframe(
            df=result,
            columns_to_check=['identity_user_id', 'user_id', 'global_contexts', 'moment'],
            check_dtypes=True,
            infer_identity_resolution=True,
        )

    @classmethod
    def anonymize_user_ids_without_identity(cls, df: bach.DataFrame) -> bach.DataFrame:
        """
        If user_id has no identity, then it will be anonymized by replacing original value with NULL.
        """
        resolved_user_id_series_name = ObjectivSupportedColumns.IDENTITY_USER_ID.value
        if resolved_user_id_series_name not in df.data_columns:
            raise Exception(
                (
                    f'Cannot anonymize user ids if {resolved_user_id_series_name} series '
                    'is not present in DataFrame data columns.'
                )
            )

        df = df.copy()
        has_no_identity = df[resolved_user_id_series_name].isnull()
        df.loc[has_no_identity, ObjectivSupportedColumns.USER_ID.value] = None
        return df

    @property
    def result_series_dtypes(self) -> Dict[str, str]:
        return {
            ObjectivSupportedColumns.USER_ID.value: bach.SeriesString.dtype,
            ObjectivSupportedColumns.IDENTITY_USER_ID.value: bach.SeriesString.dtype,
        }

    def _validate_extracted_context_df(self, df: bach.DataFrame) -> None:
        # make sure the context_df has AT LEAST the following series
        supported_dtypes = get_supported_dtypes_per_objectiv_column(
            with_identity_resolution=False,
        )
        expected_context_columns = [
            ObjectivSupportedColumns.USER_ID.value,
            ObjectivSupportedColumns.GLOBAL_CONTEXTS.value,
            ObjectivSupportedColumns.MOMENT.value,
        ]

        self._validate_data_dtypes(
            expected_dtypes={col: supported_dtypes[col] for col in expected_context_columns},
            current_dtypes=df.dtypes,
        )

    def _extract_identities_from_global_contexts(self, df: bach.DataFrame) -> bach.DataFrame:
        """
        Generates a dataframe containing the last encountered unique identity per user_id.
        This is performed by:
            1. Extract the first IdentityContext where `id` value matches the provided identity_id param value
                from the event's global_contexts. If `self._identity_id` is None, then the
                first IdentityContext found will be used instead.
            2. Drop rows where events have no IdentityContext
            3. Create the new user id based on the IdentityContext's id and name.
                Follows the following format:
                {id}|{name}
            4. Consider only the last identity from the user's last event
        returns a bach DataFrame with two series: user_id and identity_user_id.
            user-id is the user_id series from the provided `df` DataFrame
            identity_user_id is the new user_id in the format {id}|{name}
        """
        global_context_series = (
            df[ObjectivSupportedColumns.GLOBAL_CONTEXTS.value]
            .copy_override_type(bach.SeriesJson)
        )

        user_id_series_name = ObjectivSupportedColumns.USER_ID.value
        moment_series_name = ObjectivSupportedColumns.MOMENT.value

        identity_context_df = df[[user_id_series_name, moment_series_name]]

        # Extract first identity context for the event
        identity_slice_filter = {'_type': 'IdentityContext'}
        if self._identity_id is not None:
            identity_slice_filter['id'] = self._identity_id

        gc_array_slice = slice(identity_slice_filter, None)
        identity_context_df['identity_context_series'] = global_context_series.json[gc_array_slice].json[0]

        # drop events that have no identity
        identity_context_df = identity_context_df.dropna(subset=['identity_context_series'])

        # extract id and name values for the identity and create new user id
        identity_context_series = (
            identity_context_df['identity_context_series'].copy_override_type(bach.SeriesJson)
        )
        resolved_expr = bach.expression.Expression.construct(
            self.IDENTITY_FORMAT,
            identity_context_series.json.get_value('value', as_str=True),
            identity_context_series.json.get_value('id', as_str=True),
        )

        identity_user_id_series_name = ObjectivSupportedColumns.IDENTITY_USER_ID.value
        identity_context_df[identity_user_id_series_name] = (
            identity_context_df[user_id_series_name].copy_override(expression=resolved_expr)
        )

        identity_context_df = identity_context_df.materialize(node_name='extracted_id_and_name')

        # keep the last registered identity for the user
        identity_context_df = identity_context_df.drop_duplicates(
            subset=[user_id_series_name], keep='last', sort_by=moment_series_name, ascending=True,
        )
        return identity_context_df[[user_id_series_name, identity_user_id_series_name]]

    def _resolve_original_user_ids(
        self, df_to_resolve: bach.DataFrame, identity_context_df: bach.DataFrame,
    ) -> bach.DataFrame:
        """
        Replaces values from original user_id series with the extracted identities
        iff the identity is not NULL, otherwise original value is kept.
        Returns a DataFrame with updated user_ids and a new series for identity_user_id.
        """
        user_id_series_name = ObjectivSupportedColumns.USER_ID.value
        identity_user_id_series_name = ObjectivSupportedColumns.IDENTITY_USER_ID.value

        df_to_resolve = df_to_resolve.merge(
            identity_context_df, on=[user_id_series_name], how='left',
        )

        # replace user_ids with respective identity
        has_identity = df_to_resolve[identity_user_id_series_name].notnull()
        df_to_resolve.loc[has_identity, user_id_series_name] = (
            df_to_resolve[identity_user_id_series_name]
        )

        return df_to_resolve



"""
Copyright 2021 Objectiv B.V.
"""
import bach

from enum import Enum
from typing import Dict, List

from modelhub.series import series_objectiv


# Columns that Modelhub expects in an Objectiv dataframe
class ObjectivSupportedColumns(Enum):
    EVENT_ID = 'event_id'
    DAY = 'day'
    MOMENT = 'moment'
    USER_ID = 'user_id'
    GLOBAL_CONTEXTS = 'global_contexts'
    LOCATION_STACK = 'location_stack'
    EVENT_TYPE = 'event_type'
    STACK_EVENT_TYPES = 'stack_event_types'
    SESSION_ID = 'session_id'
    SESSION_HIT_NUMBER = 'session_hit_number'

    IDENTITY_USER_ID = 'identity_user_id'

    _DATA_SERIES = (
        DAY, MOMENT, USER_ID, GLOBAL_CONTEXTS, LOCATION_STACK, EVENT_TYPE,
        STACK_EVENT_TYPES, SESSION_ID, SESSION_HIT_NUMBER,
    )

    _INDEX_SERIES = (EVENT_ID, )

    _EXTRACTED_CONTEXT_COLUMNS = (
        EVENT_ID, DAY, MOMENT, USER_ID, GLOBAL_CONTEXTS, LOCATION_STACK, EVENT_TYPE, STACK_EVENT_TYPES,
    )

    _SESSIONIZED_COLUMNS = (
        SESSION_ID, SESSION_HIT_NUMBER,
    )

    _IDENTITY_RESOLUTION_COLUMNS = (IDENTITY_USER_ID, )

    @classmethod
    def get_extracted_context_columns(cls) -> List[str]:
        return list(cls._EXTRACTED_CONTEXT_COLUMNS.value)

    @classmethod
    def get_sessionized_columns(cls) -> List[str]:
        return list(cls._SESSIONIZED_COLUMNS.value)

    @classmethod
    def get_data_columns(cls) -> List[str]:
        return list(cls._DATA_SERIES.value)

    @classmethod
    def get_index_columns(cls) -> List[str]:
        return list(cls._INDEX_SERIES.value)

    @classmethod
    def get_all_columns(cls) -> List[str]:
        return cls.get_index_columns() + cls.get_data_columns()


# mapping for series names and bach series dtypes
_OBJECTIV_SUPPORTED_COLUMNS_X_SERIES_DTYPE = {
    ObjectivSupportedColumns.EVENT_ID: bach.SeriesUuid.dtype,
    ObjectivSupportedColumns.DAY: bach.SeriesDate.dtype,
    ObjectivSupportedColumns.MOMENT: bach.SeriesTimestamp.dtype,
    ObjectivSupportedColumns.USER_ID: bach.SeriesUuid.dtype,
    ObjectivSupportedColumns.GLOBAL_CONTEXTS: bach.SeriesJson.dtype,
    ObjectivSupportedColumns.LOCATION_STACK: bach.SeriesJson.dtype,
    ObjectivSupportedColumns.EVENT_TYPE: bach.SeriesString.dtype,
    ObjectivSupportedColumns.STACK_EVENT_TYPES: bach.SeriesJson.dtype,
    ObjectivSupportedColumns.SESSION_ID: bach.SeriesInt64.dtype,
    ObjectivSupportedColumns.SESSION_HIT_NUMBER: bach.SeriesInt64.dtype,

    ObjectivSupportedColumns.IDENTITY_USER_ID: bach.SeriesString.dtype,
}

# mapping for series names and modelhub series dtypes
_OBJECTIV_SUPPORTED_COLUMNS_X_MODELHUB_SERIES_DTYPE = {
    ObjectivSupportedColumns.GLOBAL_CONTEXTS: series_objectiv.SeriesGlobalContexts.dtype,
    ObjectivSupportedColumns.LOCATION_STACK: series_objectiv.SeriesLocationStack.dtype,
}


def get_supported_dtypes_per_objectiv_column(
    with_md_dtypes: bool = False, with_identity_resolution: bool = True,
) -> Dict[str, str]:
    """
    Helper function that returns mapping between Objectiv series name and dtype
    If with_md_types is true, it will return mapping against modelhub own dtypes
    If with_identity_resolution is true, the supported dtype for user_id is string
        instead of UUID.
    """
    supported_dtypes = _OBJECTIV_SUPPORTED_COLUMNS_X_SERIES_DTYPE.copy()
    if with_md_dtypes:
        supported_dtypes.update(_OBJECTIV_SUPPORTED_COLUMNS_X_MODELHUB_SERIES_DTYPE)

    if with_identity_resolution:
        supported_dtypes.update({ObjectivSupportedColumns.USER_ID: bach.SeriesString.dtype})

    return {col.value: dtype for col, dtype in supported_dtypes.items()}


def check_objectiv_dataframe(
    df: bach.DataFrame,
    columns_to_check: List[str] = None,
    check_index: bool = False,
    check_dtypes: bool = False,
    with_md_dtypes: bool = False,
    infer_identity_resolution: bool = True,
) -> None:
    """
    Helper function that determines if provided dataframe is an objectiv dataframe.
    :param df: bach DataFrame to be checked
    :param columns_to_check: list of columns to verify,
        if not provided, all expected objectiv columns will be used instead.
    :param check_index: if true, will check if dataframe has expected index series
    :param check_dtypes: if true, will check if each series has expected dtypes
    :param with_md_dtypes: if true, will check if series has expected modelhub dtype
    :param infer_identity_resolution: if true, will check if user_id series has string dtype
    """
    columns = columns_to_check if columns_to_check else ObjectivSupportedColumns.get_all_columns()

    with_identity_resolution = False
    if infer_identity_resolution:
        if ObjectivSupportedColumns.IDENTITY_USER_ID.value in df.all_series:
            # if df has identity_user_id series, we can infer user_id series was changed
            # by IdentityResolutionPipeline
            with_identity_resolution = True
        elif (
            ObjectivSupportedColumns.USER_ID.value in df.all_series
            and df.all_series[ObjectivSupportedColumns.USER_ID.value].dtype == bach.SeriesString.dtype
        ):
            # user_id is dtype string, it was changed by IdentityResolutionPipeline
            with_identity_resolution = True

    supported_dtypes = get_supported_dtypes_per_objectiv_column(
        with_md_dtypes=with_md_dtypes, with_identity_resolution=with_identity_resolution,
    )

    for col in columns:
        if col not in supported_dtypes:
            raise ValueError(f'{col} is not present in Objectiv supported columns.')

        if col not in df.all_series:
            raise ValueError(f'{col} is not present in DataFrame.')

        if (
            check_index
            and col in ObjectivSupportedColumns.get_index_columns()
            and col not in df.index
        ):
            raise ValueError(f'{col} is not present in DataFrame index.')

        if check_dtypes:
            dtype = supported_dtypes[col]
            if df.all_series[col].dtype != dtype:
                raise ValueError(f'{col} must be {dtype} dtype.')



from typing import Optional

from sqlalchemy.engine import Engine

from modelhub.util import ObjectivSupportedColumns, check_objectiv_dataframe

import bach


def get_objectiv_data(
    *,
    engine: Engine,
    table_name: str,
    session_gap_seconds: int = 1800,
    set_index: bool = True,
    start_date: Optional[str] = None,
    end_date: Optional[str] = None,
    with_sessionized_data: bool = True,
    identity_resolution: Optional[str] = None,
    anonymize_unidentified_users: bool = True,
) -> bach.DataFrame:
    """
        :param engine: db_connection
        :param table_name: the name of the sql table where the data is stored. Will default to 'events' for
            bigquery and 'data' for other engines.
        :param start_date: start_date to filter data from ExtractedContextsPipeline
        :param end_date: end_date to filter data from ExtractedContextsPipeline
        :param with_sessionized_data: If true, SessionizedDataPipeline will be applied on
            ExtractedContextsPipeline (or IdentityResolutionPipeline) result.
        :param session_gap_seconds: Maximum time between two consecutive
            events from the same user, for the events to be considered part of the same session.
            This is used by SessionizedDataPipeline, thus only relevant if `with_sessionized_data=True`.
        :param identity_resolution: If value provided, IdentityResolutionPipeline will be applied
            on ExtractedContextsPipeline result
        :param anonymize_unidentified_users: If True, unidentified user_ids will be set to NULL.
            This step is performed after applying IdentityResolutionPipeline and SessionizedDataPipeline.

        :returns: initial bach DataFrame required by ModelHub.
    """
    from modelhub.pipelines import (
        ExtractedContextsPipeline, SessionizedDataPipeline, IdentityResolutionPipeline
    )

    contexts_pipeline = ExtractedContextsPipeline(engine=engine, table_name=table_name)
    sessionized_pipeline = SessionizedDataPipeline(session_gap_seconds=session_gap_seconds)
    identity_pipeline = IdentityResolutionPipeline(identity_id=identity_resolution)

    data = contexts_pipeline(start_date=start_date, end_date=end_date)

    # resolve user ids
    if identity_resolution:
        data = identity_pipeline(extracted_contexts_df=data)

    # calculate sessionized data from events
    if with_sessionized_data:
        data = sessionized_pipeline(extracted_contexts_df=data)

    # Anonymizing users must be done after getting sessionized data, this way we don't aggregate
    # events on a single unknown user
    if identity_resolution and anonymize_unidentified_users:
        data = IdentityResolutionPipeline.anonymize_user_ids_without_identity(data)

    columns_to_check = ObjectivSupportedColumns.get_extracted_context_columns()
    if with_sessionized_data:
        columns_to_check += ObjectivSupportedColumns.get_sessionized_columns()

    check_objectiv_dataframe(
        df=data,
        columns_to_check=columns_to_check,
        check_dtypes=True,
        infer_identity_resolution=identity_resolution is not None,
    )
    data = data[columns_to_check]

    if set_index:
        data = data.set_index(keys=ObjectivSupportedColumns.get_index_columns())

    return data



import json
from datetime import datetime, timezone
from typing import Dict, Any, NamedTuple, Optional
from uuid import UUID

import bach
from bach import DataFrame
from sql_models.constants import DBDialect
from sql_models.util import is_postgres, is_bigquery
from sqlalchemy import create_engine
from sqlalchemy.engine import Engine
from tests.functional.bach.test_data_and_utils import run_query
from tests.unit.bach.util import get_pandas_df

from modelhub import ModelHub
from tests_modelhub.data_and_utils.data_json_real import TEST_DATA_JSON_REAL, JSON_COLUMNS_REAL
from tests_modelhub.data_and_utils.data_objectiv import TEST_DATA_OBJECTIV


class DBParams(NamedTuple):
    url: str
    credentials: Optional[str]
    table_name: str


def _convert_moment_to_utc_time(moment: str) -> int:
    dt = datetime.fromisoformat(moment)
    dt = dt.replace(tzinfo=timezone.utc)
    return int(dt.timestamp() * 1e3)


def get_df_with_json_data_real(db_params: DBParams) -> DataFrame:
    engine = create_engine_from_db_params(db_params)
    df = DataFrame.from_pandas(
        engine=engine,
        df=get_pandas_df(dataset=TEST_DATA_JSON_REAL, columns=JSON_COLUMNS_REAL),
        convert_objects=True,
    )
    df['global_contexts'] = df.global_contexts.astype('json')
    df['location_stack'] = df.location_stack.astype('json')
    return df


def get_objectiv_dataframe_test(db_params=None, time_aggregation=None):
    if not db_params:
        # by default use PG (this should be removed after modelhub is able to work with all bach engines)
        import os
        db_url = os.environ.get('OBJ_DB_PG_TEST_URL', 'postgresql://objectiv:@localhost:5432/objectiv')
        credentials = None
        table_name = 'objectiv_data'
    else:
        db_url = db_params.url
        credentials = db_params.credentials
        table_name = db_params.table_name

    kwargs = {}
    if time_aggregation:
        kwargs = {'time_aggregation': time_aggregation}
    modelhub = ModelHub(**kwargs)

    return modelhub.get_objectiv_dataframe(
        db_url=db_url,
        table_name=table_name,
        bq_credentials_path=credentials,
    ), modelhub


def get_parsed_objectiv_data(engine):
    parsed_data = []
    for event_data in TEST_DATA_OBJECTIV:
        event_id, day, moment, cookie_id, value = event_data
        value = json.loads(value)
        # BQ uses time from taxonomy json for getting moment and day
        # therefore time value MUST be the same as moment
        if is_bigquery(engine):
            value['time'] = _convert_moment_to_utc_time(moment)

        parsed_data.append(
            {
                'event_id': UUID(event_id),
                'day': datetime.strptime(day, '%Y-%m-%d').date(),
                'moment': datetime.fromisoformat(moment),
                'cookie_id': UUID(cookie_id),
                'value': value
            }
        )

    return parsed_data


def create_engine_from_db_params(db_params: DBParams) -> Engine:
    if db_params.credentials:
        engine = create_engine(url=db_params.url, credentials_path=db_params.credentials)
    else:
        engine = create_engine(url=db_params.url)

    return engine


def setup_db(engine: Engine, table_name: str):
    columns = {
        'event_id': bach.SeriesUuid.supported_db_dtype[DBDialect.POSTGRES],
        'day': bach.SeriesDate.supported_db_dtype[DBDialect.POSTGRES],
        'moment': bach.SeriesTimestamp.supported_db_dtype[DBDialect.POSTGRES],
        'cookie_id': bach.SeriesUuid.supported_db_dtype[DBDialect.POSTGRES],
        'value': bach.SeriesJson.supported_db_dtype[DBDialect.POSTGRES],
    }
    _prep_db_table(engine, table_name=table_name, columns=columns)
    _insert_records_in_db(engine, table_name=table_name, columns=columns)


def _prep_db_table(engine, table_name: str, columns: Dict[str, Any]):
    if is_postgres(engine):
        column_stmt = ','.join(f'{col_name} {db_type}' for col_name, db_type in columns.items())
        sql = f"""
            drop table if exists {table_name};
            create table {table_name} ({column_stmt});
            alter table {table_name}
                owner to objectiv
        """
    else:
        raise Exception()
    run_query(engine, sql)


def _insert_records_in_db(engine, table_name: str, columns: Dict[str, Any]):
    from tests_modelhub.data_and_utils.data_objectiv import TEST_DATA_OBJECTIV

    column_stmt = ','.join(columns.keys())
    records = []
    if is_postgres(engine):
        for record in TEST_DATA_OBJECTIV:
            formatted_values = [f"'{record[col_index]}'" for col_index, _ in enumerate(columns)]
            records.append(f"({','.join(formatted_values)})")
    else:
        raise Exception()

    values_stmt = ','.join(records)
    sql = f'insert into {table_name} ({column_stmt}) values {values_stmt}'
    run_query(engine, sql)



"""
Copyright 2021 Objectiv B.V.
"""
import re
from typing import List, Union, Dict, Tuple, Optional, cast
from typing import TYPE_CHECKING

import bach
from sqlalchemy import create_engine
from sqlalchemy.engine import Engine

from modelhub.aggregate import Aggregate
from modelhub.map import Map
from modelhub.models.logistic_regression import LogisticRegression
from modelhub.series.series_objectiv import MetaBase
from sql_models.constants import NotSet
from sql_models.util import is_bigquery

if TYPE_CHECKING:
    from modelhub.series import SeriesLocationStack

GroupByType = Union[List[Union[str, bach.Series]], str, bach.Series, NotSet]
ConversionEventDefinitionType = Tuple[Optional['SeriesLocationStack'], Optional[str]]

TIME_DEFAULT_FORMAT = '%Y-%m-%d %H:%M:%S.%f'
SESSION_GAP_DEFAULT_SECONDS = 1800


class ModelHub:
    """
    The model hub contains collection of data models and convenience functions that you can take, combine and
    run on Bach data frames to quickly build highly specific model stacks for product analysis and
    exploration.
    It includes models for a wide range of typical product analytics use cases.

    All models from the model hub can run on Bach DataFrames that contain data collected by the Objectiv
    tracker. To instantiate a DataFrame with Objectiv data use :py:meth:`get_objectiv_dataframe`. Models
    from the model hub assume that at least the columns of a DataFrame instantiated with this method are
    available in order to run properly. These columns are:

    The model hub has three main type of functions: helper functions, aggregation models and
    machine learning models.

    * Helper functions always return a series with the same shape and index as the DataFrame they originate
      from. This ensures they can be added as a column to that DataFrame. The helper functions can be accessed
      with the :py:attr:`map` accessor from a model hub instance.
    * Aggregation models return aggregated data in some form from the DataFrame. The aggregation models can be
      accessed with the :py:attr:`agg` or :py:attr:`aggregate` accessor from a model hub instance.
    * Machine learning models can be instantiated from the modelhub directly using the model's name,
      i.e. : :py:meth:`get_logistic_regression`.
    """
    def __init__(self,
                 time_aggregation: str = TIME_DEFAULT_FORMAT):
        """
        Constructor

        :param time_aggregation: Time aggregation used for aggregation models.
        """

        self._time_aggregation = time_aggregation
        self._conversion_events = cast(Dict[str, ConversionEventDefinitionType], {})

        # init metabase
        self._metabase = None

    @property
    def time_aggregation(self):
        """
        Time aggregation used for aggregation models, set when object is instantiated.
        """
        return self._time_aggregation

    @property
    def conversion_events(self):
        """
        Dictionary of all events that are labeled as conversion.

        Set with :py:meth:`add_conversion_event`
        """
        return self._conversion_events

    @staticmethod
    def _get_db_engine(db_url: Optional[str], bq_credentials_path: Optional[str] = None) -> Engine:
        """
        returns db_connection based on db_url.
        If db_url is for BigQuery, bq_credentials_path must be provided.
        """
        if db_url and re.match(r'^bigquery://.+', db_url):
            if not bq_credentials_path:
                raise ValueError('BigQuery credentials path is required for engine creation.')

            return create_engine(db_url, credentials_path=bq_credentials_path)

        import os
        db_url = db_url or os.environ.get('DSN', 'postgresql://objectiv:@localhost:5432/objectiv')
        return create_engine(db_url)

    def get_objectiv_dataframe(
        self,
        *,
        db_url: str = None,
        table_name: str = None,
        start_date: str = None,
        end_date: str = None,
        bq_credentials_path: Optional[str] = None,
        with_sessionized_data: bool = True,
        session_gap_seconds: int = SESSION_GAP_DEFAULT_SECONDS,
        identity_resolution: Optional[str] = None,
        anonymize_unidentified_users: bool = True,
    ):
        """
        Sets data from sql table into an :py:class:`bach.DataFrame` object.

        The created DataFrame points to where the data is stored in the sql database, makes several
        transformations and sets the right data types for all columns. As such, the models from the model hub
        can be applied to a DataFrame created with this method.

        :param db_url: the url that indicate database dialect and connection arguments. If not given, env DSN
            is used to create one. If that's not there, the default of
            'postgresql://objectiv:@localhost:5432/objectiv' will be used.
        :param table_name: the name of the sql table where the data is stored. Will default to 'events' for
            bigquery and 'data' for other engines.
        :param start_date: first date for which data is loaded to the DataFrame. If None, data is loaded from
            the first date in the sql table. Format as 'YYYY-MM-DD'.
        :param end_date: last date for which data is loaded to the DataFrame. If None, data is loaded up to
            and including the last date in the sql table. Format as 'YYYY-MM-DD'.
        :param bq_credentials_path: path for BigQuery credentials. If db_url is for BigQuery engine, this
            parameter is required.
        :param with_sessionized_data: Indicates if DataFrame must include `session_id`
            and `session_hit_number` calculated series.
        :param session_gap_seconds: Amount of seconds to be use for identifying if events were triggered
            or not during the same session.
        :param identity_resolution: Identity id to be used for identifying users based on IdentityContext.
            If no value is provided, then the user_id series will contain the value from
            the cookie_id column (a UUID).
        :param anonymize_unidentified_users: Indicates if unidentified users are required to be anonymized
            by setting user_id value to NULL. Otherwise, original UUID value from the cookie will remain.

        :returns: :py:class:`bach.DataFrame` with Objectiv data.

        .. note::
            If `with_sessionized_data` is True, Objectiv data will include `session_id` (int64)
                and `session_hit_number` (int64) series.
        """
        engine = self._get_db_engine(db_url=db_url, bq_credentials_path=bq_credentials_path)
        from modelhub.pipelines.util import get_objectiv_data
        if table_name is None:
            if is_bigquery(engine):
                table_name = 'events'
            else:
                table_name = 'data'

        data = get_objectiv_data(
            engine=engine,
            table_name=table_name,
            start_date=start_date,
            end_date=end_date,
            with_sessionized_data=with_sessionized_data,
            session_gap_seconds=session_gap_seconds,
            identity_resolution=identity_resolution,
            anonymize_unidentified_users=anonymize_unidentified_users,
        )

        # get_objectiv_data returns both series as bach.SeriesJson.
        data['global_contexts'] = data.global_contexts.astype('objectiv_global_context')
        data['location_stack'] = data.location_stack.astype('objectiv_location_stack')
        return data

    def add_conversion_event(self,
                             location_stack: 'SeriesLocationStack' = None,
                             event_type: str = None,
                             name: str = None):
        """
        Label events that are used as conversions. All labeled conversion events are set in
        :py:attr:`conversion_events`.

        :param location_stack: the location stack that is labeled as conversion. Can be any slice in of a
            :py:class:`modelhub.SeriesLocationStack` type column. Optionally use in conjunction with
            ``event_type`` to label a conversion.
        :param event_type: the event type that is labeled as conversion. Optionally use in conjunction with
            ``objectiv_location_stack`` to label a conversion.
        :param name: the name to use for the labeled conversion event. If None it will use 'conversion_#',
            where # is the number of the added conversion.
        """

        if location_stack is None and event_type is None:
            raise ValueError('At least one of conversion_stack or conversion_event should be set.')

        if not name:
            name = f'conversion_{len(self._conversion_events) + 1}'

        self._conversion_events[name] = location_stack, event_type

    def time_agg(self, data: bach.DataFrame, time_aggregation: str = None) -> bach.SeriesString:
        """
        Formats the moment column in the DataFrame, returns a SeriesString.

        Can be used to aggregate to different time intervals, ie day, month etc.

        :param data: :py:class:`bach.DataFrame` to apply the method on.
        :param time_aggregation: if None, it uses :py:attr:`time_aggregation` set from the
            ModelHub. Use any template for aggregation based on 1989 C standard format codes:
            https://docs.python.org/3/library/datetime.html#strftime-and-strptime-format-codes
            ie. ``time_aggregation=='%Y-%m-%d'`` aggregates by date.
        :returns: SeriesString.
        """

        time_aggregation = self.time_aggregation if time_aggregation is None else time_aggregation
        return data.moment.dt.strftime(time_aggregation).copy_override(name='time_aggregation')

    _metabase: Union[None, MetaBase] = None

    def to_metabase(self, data, model_type: str = None, config: dict = None):
        """
        Plot data in ``data`` to Metabase. If a card already exists, it will be updated. If ``data`` is a
        :py:class:`bach.Series`, it will call :py:meth:`bach.Series.to_frame`.

        Default options can be overridden using the config dictionary.

        :param data: :py:class:`bach.DataFrame` or :py:class:`bach.Series` to push to MetaBase.
        :param model_type: Preset output to Metabase for a specific model. eg, 'unique_users'
        :param config: Override default config options for the graph to be added/updated in Metabase.
        """
        if not self._metabase:
            self._metabase = MetaBase()
        return self._metabase.to_metabase(data, model_type, config)

    @property
    def map(self):
        """
        Access map methods from the model hub.

        .. autoclass:: Map
            :members:
            :noindex:

        """

        return Map(self)

    @property
    def agg(self):
        """
        Access aggregation methods from the model hub. Same as :py:attr:`aggregate`.

        .. autoclass:: Aggregate
            :members:
            :noindex:

        """

        return Aggregate(self)

    @property
    def aggregate(self):
        """
        Access aggregation methods from the model hub. Same as :py:attr:`agg`.

        .. autoclass:: Aggregate
            :members:
            :noindex:

        """
        return Aggregate(self)

    def get_logistic_regression(self, *args, **kwargs) -> LogisticRegression:
        """
        Return an instance of the :py:class:`LogisticRegression` from the model hub.

        All parameters passed to this function are passed to the constructor of the LogisticRegression
        model.
        """

        return LogisticRegression(*args, **kwargs)



"""
Copyright 2021 Objectiv B.V.
"""
__version__ = '0.0.17'

from modelhub.modelhub import ModelHub
from modelhub.aggregate import Aggregate
from modelhub.map import Map
from modelhub.models.logistic_regression import LogisticRegression
from modelhub.pipelines import *
from modelhub.series import *


# Here we do a basic version check, to make sure we are on the most recent versions of objectiv-bach and
# objectiv-modelhub. This is done by querying the backend that holds a cached version of the latest versions
# available from the pypi archive. These are compared with the local versions. If a newer version is available
# a Python warning is issued.
# To disable this, either set `OBJECTIV_VERSION_CHECK_DISABLE` in the environment, or suppress the warning.
#
# See: https://objectiv.io/docs/modeling/open-model-hub/version-check/` for more info

# we need this to check the environment variables
import os


def check_version():
    # check env for opt-out setting
    if os.environ.get('OBJECTIV_VERSION_CHECK_DISABLE', 'false') != 'false':
        return
    try:
        # wrap the import in try/except to make sure we don't fail if there are missing imports
        import warnings
        import requests
        from bach import __version__ as bach_version

        check_url = os.environ.get('OBJECTIV_VERSION_CHECK_URL',
                                   'https://version-check.objectiv.io/check_version')
        packages = [
                f'objectiv-bach:{bach_version}',
                f'objectiv-modelhub:{__version__}'
        ]
        data = '\n'.join(packages)

        response = requests.post(check_url, data=data, timeout=5)
        lines = response.text
        for line in lines.split('\n'):
            items = line.split(':')
            # we expect at least 4 items, but the message may contain colons, so there
            # may be more items in the list. We combine the remaining ones into
            # one str: message
            if len(items) > 3:
                package, updated, version = items[:3]
                message = ':'.join(items[3:])
                # this is a line containing package:updated:version:message

                if updated == 'True':
                    warnings.warn(category=Warning, message=message)
    except Exception as e:
        pass


check_version()



"""
Copyright 2021 Objectiv B.V.
"""

# Any import from modelhub initializes all the types, do not remove



from typing import Iterable
import bach
from sklearn.linear_model import LogisticRegression
from modelhub import ModelHub
import numpy as np


class TestLR:
    """
    Tests if model used on Bach gives the same outcome as model used on sklearn.
    Note that it only works for solvers that provide consistent outcomes.
    """

    def __init__(self, X: bach.DataFrame, y: bach.Series, **kwargs):
        """
        :param X: the independent variables.
        :param y: the target variable.
        :param **kwargs: any parameters that sklearn's LogisticRegression takes for instantiating the model.
            These parameters are used also for the model hub version of Logistic Regression.
        """

        self.X = X
        self.y = y

        data = X.copy()
        data[y.name] = y
        pdf = data.to_pandas()

        self.X_p = pdf[X.data_columns]
        self.y_p = pdf[y.name]

        modelhub = ModelHub()

        self.sklearn_lr = LogisticRegression(**kwargs)
        self.modelhub_lr = modelhub.get_logistic_regression(**kwargs)

        self.sklearn_lr.fit(self.X_p, self.y_p)
        self.modelhub_lr.fit(self.X, self.y)

    def test_fitted_model(self):
        for key, value in self.sklearn_lr.__dict__.items():
            modelhub_value = getattr(self.modelhub_lr, key)
            print(f'testing {key}')
            print(f'modelhub value: {modelhub_value}\nsklearn value : {value}\n')
            result = modelhub_value == value
            if isinstance(result, Iterable):
                result = result.all()
            assert result

    def test_method(self, method_name, X=False, y=False, **kwargs):
        """
        tests if modelhub outcome of method is the same as sklearn.
        """

        if method_name not in ['decision_function',
                               'predict',
                               'predict_proba',
                               'score']:
            raise NotImplementedError(f"method {method_name} not supported")

        sklearn_args = tuple()
        modelhub_args = tuple()

        if X:
            # need sorting for the comparison
            sklearn_args = (self.X_p.sort_index(),)
            modelhub_args = (self.X.sort_index(),)
            if y:
                # for now, we only calculate score, sorting is not needed
                sklearn_args = (self.X_p, self.y_p)
                modelhub_args = (self.X, self.y)

        sklearn_method = getattr(self.sklearn_lr, method_name)
        modelhub_method = getattr(self.modelhub_lr, method_name)
        sklearn_data = sklearn_method(*sklearn_args, **kwargs)
        modelhub_data = modelhub_method(*modelhub_args, **kwargs)

        if method_name == 'predict_proba':
            # extract the probabilities from sklearn's predicted probabilities for the True class only:
            sklearn_data = sklearn_data[:, np.where(self.sklearn_lr.classes_)[0][0]]

        if method_name == 'score':
            equals = np.isclose(sklearn_data, modelhub_data)
        else:
            modelhub_data = modelhub_data.sort_index().to_numpy()
            equals = np.isclose(sklearn_data, modelhub_data).all()

        assert equals, f"modelhub_data: {modelhub_data} != sklearn_data: {sklearn_data}"

        print(f"test ok")
        print(f'modelhub value: {modelhub_data}\nsklearn value : {sklearn_data}\n')




"""
Copyright 2022 Objectiv B.V.

In this module we setup the database for functional tests.
"""
from pathlib import Path

from filelock import FileLock
import pytest
from _pytest.fixtures import SubRequest
from _pytest.tmpdir import TempPathFactory
from sqlalchemy import create_engine

from tests_modelhub.conftest import get_postgres_db_params
from tests_modelhub.data_and_utils.utils import setup_db


@pytest.fixture(autouse=True, scope='session')
def setup_postgres_db(request: SubRequest, tmp_path_factory: TempPathFactory, worker_id: str) -> None:
    """
    Helper for creating postgres database used by all functional tests. Only created if it is required
    to run tests against Postgres.
    When running in a scenario with multiple test processes, this will use a filelock to ensure the database
    get created only once.
    """
    # use same logic as tests_modelhub.conftest.pytest_generate_tests()
    testing_pg = not request.session.config.getoption("big_query")
    if not testing_pg:
        return

    if worker_id == 'master':
        # This is the simple case: we are running with a single worker
        _real_setup_postgres_db()
        return

    # We are running with multiple workers. Make sure only one worker calls _real_setup_postgres_db()
    # Use a FileLock, as advised by the pytest-xdist docs [1]
    # [1] https://pypi.org/project/pytest-xdist/#making-session-scoped-fixtures-execute-only-once
    # Basic idea is that we'll use a lock file to make sure only one process at a time is in the critical
    # section. Besides setting up the database in the critical section, we'll also create a file on disk to
    # indicate to the other processes that the database has been setup already.

    root_tmp_dir = tmp_path_factory.getbasetemp().parent  # get the temp directory shared by all workers
    is_done_path: Path = root_tmp_dir / "setup_database_is_done.txt"
    lock_path: Path = root_tmp_dir / "setup_database.lock"

    # Try to get lock, but give up after 10 seconds. If another process is still busy setting up the database
    # after 10 seconds, then something is likely wrong. Better to fail than to wait forever.
    lock = FileLock(str(lock_path), timeout=10)
    with lock:
        if not is_done_path.is_file():
            # We got the lock, but the file does not yet exist, so we are the first process
            _real_setup_postgres_db()
            is_done_path.write_text('done')
        # else case: We got the lock, but the is_done_path file already exists, indicating we don't have to
        # do anything.


def _real_setup_postgres_db():
    db_params = get_postgres_db_params()
    setup_db(engine=create_engine(url=db_params.url), table_name=db_params.table_name)



"""
Copyright 2021 Objectiv B.V.
"""

# Any import from modelhub initializes all the types, do not remove


import pytest
from tests_modelhub.data_and_utils.utils import get_objectiv_dataframe_test
from tests_modelhub.functional.modelhub.logistic_regression_test_utils import TestLR


def test_fitted_model(db_params):
    bt, modelhub = get_objectiv_dataframe_test(db_params)
    bt['ts'] = bt['moment'].dt.strftime('%m%d%H%m').astype(dtype=int)
    bt['session'] = bt['session_hit_number'] * 100
    bt['target'] = bt.session_hit_number > 1

    test_lr = TestLR(X=bt[['ts', 'session_hit_number']],
                     y=bt['target'])

    test_lr.test_fitted_model()


@pytest.mark.parametrize("method_name,X,y", [
    ('predict', True, False),
    ('predict_proba', True, False),
    ('score', True, True),
])
def test_model_methods(db_params, method_name, X, y):
    bt, modelhub = get_objectiv_dataframe_test(db_params)
    bt['ts'] = bt['moment'].dt.strftime('%m%d%H%m').astype(dtype=int)
    bt['session'] = bt['session_hit_number'] * 100
    bt['target'] = bt.session_hit_number > 1

    test_lr = TestLR(X=bt[['ts', 'session_hit_number', 'session']],
                     y=bt['target'])

    test_lr.test_method(method_name=method_name, X=X, y=y)
